package vacworld;

import agent.*;
import java.io.*;

/** A simulator for the vacuum cleaning world environment. This environment
    is inaccessible, deterministic, static and discrete. */
public class VacuumWorld extends Environment{

  // scoring information 
  protected int numMoves = 0;
  protected int numTurns = 0;
  protected int numSucks = 0;
  protected int numBumps = 0;
  protected int totalScore = 0;

  protected PrintStream output;
  protected boolean interactive;

  public VacuumWorld() {
    output = System.out;
    interactive = true;
  }

  public VacuumWorld(PrintStream output, boolean interactive) {

    this.output = output;
    this.interactive = interactive;
  }

  /** Add a new agent to the environment. Since the vacuum cleaning
      world is a single agent environment, this method should only
      be called once per object. */
  public void addAgent(Agent agent) {

    if (agents.isEmpty() == false) {
      output.println("ERROR - tried to add a second agent to a single agent environment");
    }
    else {
      agents.add(agent);
    }
  }
 

  /** The simulation is complete when the robot has performed its ShutOff
      action. */
  protected boolean isComplete() {

    if (((VacuumState)state).isRobotOff())
      return true;
    else
      return false;
  }

  
  /** Create a percept for an agent. This implements the see: E -> Per
      function. */
  protected Percept getPercept(Agent a) {
    
    VacPercept p;

    if (state instanceof VacuumState) {
      p = new VacPercept((VacuumState)state, a);
      output.println("Pecept: " + p.toString());
      return p;
    }
    else {
      output.println("ERROR - state is not a VacuumState object.");
      return null;
    }
  }

  
  /** Execute an agent's action and update the environment's state. */
  protected void updateState(Agent a, Action action) {

    super.updateState(a, action);
    
    // update scoring data
    if (action instanceof GoForward)
      numMoves++;
    else if (action instanceof TurnRight)
      numTurns++;
    else if (action instanceof SuckDirt)
      numSucks++;
    if (((VacuumState)state).bumped())
      numBumps++;

    output.println("Action: " + action.toString());
    output.println();
  }


  /** Starts the simulation. We override the method in agent.Environment so
      that VacuumWorld states can be displayed to a selected output stream. */
  public void start(State initState) {

    Percept p;
    Action action;
    Agent a;

    state = initState;
    ((VacuumState)state).display(output);
    if (interactive == true)
      waitForUser();

    a = (Agent)agents.firstElement();
    while (!isComplete()) {
      p = getPercept(a);
      a.see(p);
      action = a.selectAction();
      updateState(a, action);
      ((VacuumState)state).display(output);
      if (interactive == true)
	waitForUser();
    }
    totalScore = getPerformanceMeasure();
  }


  /** Pause simulation until user has pressed a key. */
  protected void waitForUser() {
      
    String input;

    System.out.println("Press ENTER to continue.");
    
    BufferedReader console = new BufferedReader(new InputStreamReader(System.in));
    try {
      input = console.readLine();
    }
    catch (IOException e) {
      System.out.println(e.getMessage());
      return;
    }
  }

  /** Runs the program. */
  public static void main(String[] args) {

    VacuumWorld world;
    VacAgent agent;
    VacuumState initState;

    System.out.println("The Vacuum Cleaner World Agent Testbed");
    System.out.println("--------------------------------------");
    System.out.println();
    
    world = new VacuumWorld();
    agent = new VacAgent();
    initState = VacuumState.getInitState();
    // initState = VacuumState.getRandomState();
    world.addAgent(agent);
    world.start(initState);

    System.out.println();
    world.printScore(System.out);
  }

  protected int getNumDirtyLocs() {

    return ((VacuumState)state).getNumDirtyLocs();
  }

  /** Returns true if the robot has turned itself off in its origin
      square. */
  public boolean gotHome() {
    
    VacuumState vstate;

    vstate = (VacuumState)state;
    if (((VacuumState)state).isRobotOff() && 
        vstate.getAgentX() == 1 && vstate.getAgentY() == 1)
      return true;
    else
      return false;
  }

  /** Returns the number of actions the agent has executed. */
  protected int getNumActions() {

    return numMoves + numTurns + numSucks;
  }

  /** Returns the performance measure of the agent in the current 
      environment. */
  protected int getPerformanceMeasure() {

    int score;
    score = 1000 + getMovesScore() + getTurnsScore() + getSucksScore() +
	getBumpsScore() + getDirtScore() + getHomeScore();
    if (score < 0)
      score = 0;
    return score;
  }

  private int getMovesScore() {
    return numMoves * -2;
  }

  private int getTurnsScore() {
    return numTurns * -1;
  }

  private int getSucksScore() {
    return numSucks * -2;
  }

  private int getBumpsScore() {
    return numBumps * -10;
  }

  private int getDirtScore() {
    return getNumDirtyLocs() * -100;
  }

  private int getHomeScore() {
    if (gotHome() == false)
      return -200;
    else
      return 0;
  }

  protected void printScore(PrintStream out) {

    out.println("Evaluation:");
    out.println("----------");
    out.println("  Base score: \t1000");
    out.println("  Home penalty: " + getHomeScore());
    out.println("  Dirt left: \t" + getDirtScore() + "\t(" + 
                getNumDirtyLocs() + ")");
    out.println("  Bumps: \t" + getBumpsScore() + "\t(" + numBumps + ")");
    out.println("  Moves: \t" + getMovesScore() + "\t(" + numMoves + ")");
    out.println("  Turns: \t" + getTurnsScore() + "\t(" + numTurns + ")");
    out.println("  Sucks: \t" + getSucksScore() + "\t(" + numSucks + ")");
    out.println("  ----------------------------");
    out.println("  TOTAL SCORE: \t" + totalScore);
    out.println();
  }


}
